#include <modules.h>

gint gmodule_compare_func(gconstpointer a,gconstpointer b) {

  if(g_strcasecmp(g_module_name((GModule*)a),
		  g_module_name((GModule*)b)) == 0) return TRUE;
  return FALSE;

}

gboolean gmodule_remove_func(gpointer key,gpointer value,gpointer user_data) {

  g_print("closing (%s)\n",(gchar*)key);

  g_module_close((GModule*)value);
  g_free((gchar*)key);

  /* true in this case means remove the module from the hash,
   * no questions asked...
   */
  return TRUE;

}

GHashTable *module_hash_new(void) {
  
  GHashTable *new =
	  g_hash_table_new(g_str_hash,gmodule_compare_func);

  return new;

}

void module_hash_destroy(GHashTable *hash) {

  if(hash != NULL) {
    gint n;
    n = g_hash_table_foreach_remove(hash,
				    gmodule_remove_func,
				    NULL);
    g_print("%d modules removed\n",n);
    g_hash_table_destroy(hash);
    hash = NULL;
  }
  
}

gboolean call_func(GHashTable *hash,
		   const gchar *dir,const gchar *lib,const gchar *func,
		   const gint argc, const gpointer argv[]) {
  
  GModule *mod;
  gboolean (*funcptr)(const gint argc,const gpointer argv[]);
  gboolean rv = FALSE;
  gchar *str = NULL;

  if(!g_module_supported()) {
    g_error("modules are not supported, this program needs module support.\n");
    /* no return, because g_error() calls abort() */
  }

  str = g_module_build_path(dir,lib);

  if(hash!=NULL) {

    if((mod=g_hash_table_lookup(hash,str))==NULL) {
      
      mod = g_module_open(str,G_MODULE_BIND_LAZY);
      if(mod==NULL) {
	g_warning(g_module_error());
	return FALSE;
      }

      g_hash_table_insert(hash,g_strdup(str),mod);
    
    }
    
    if(!g_module_symbol(mod,func,(gpointer)&funcptr)) {
      g_warning(g_module_error());
      return FALSE;
    }

    rv = (*funcptr)(argc,argv);

  } else { /* if(hash==NULL) */
    
    mod = g_module_open(str,G_MODULE_BIND_LAZY);
    if(mod==NULL) {
      g_warning(g_module_error());
      return FALSE;
    }

    if(!g_module_symbol(mod,func,(gpointer)&funcptr)) {
      g_warning(g_module_error());
      return FALSE;
    }

    rv = (*funcptr)(argc,argv);

    g_module_close(mod);

  }

  if(str!=NULL) g_free(str);

  return rv;

}
